require 'heartland_portico'

module ActiveMerchant #:nodoc:
  module Billing #:nodoc:
    class HeartlandPorticoGateway < Gateway
      #self.test_url = 'https://example.com/test'
      #self.live_url = 'https://example.com/live'

      # The countries the gateway supports merchants from as 2 digit ISO country codes
      self.supported_countries = ['US']

      # The card types supported by the payment gateway
      self.supported_cardtypes = [:visa, :master, :american_express, :discover]

      # The homepage URL of the gateway
      self.homepage_url = 'http://www.heartlandpaymentsystems.com/Payment-Processing/Payment-Processing/Portico-Virtual-Terminal'

      # The name of the gateway
      self.display_name = 'Heartland Portico'

      self.money_format = :dollars

      def initialize(options = {})
        @options = options
        requires!(@options, :user_name, :password)
        requires!(@options, :developer_i_d, :device_id, :license_id, :site_id, :version_nbr)
        super
      end

      def verify(creditcard, options = {})
        post = {}
        add_creditcard(post, creditcard)
        add_address(post, creditcard, options)
        add_customer_data(post, options)

        commit('credit_account_verify', nil, post)
      end

      def authorize(money, creditcard, options = {})
        post = initial_post(options)
        add_invoice(post, options)
        add_creditcard(post, creditcard)
        add_address(post, creditcard, options)
        add_customer_data(post, options)

        commit('credit_auth', money, post)
      end

      def purchase(money, creditcard, options = {})
        post = initial_post(options)
        add_invoice(post, options)
        add_creditcard(post, creditcard)
        add_address(post, creditcard, options)
        add_customer_data(post, options)

        commit('credit_sale', money, post)
      end

      def capture(money, authorization, options = {})
        post = initial_post(options)
        add_authorization(post, authorization)

        commit('credit_add_to_batch', money, post)
      end

      def return(money, creditcard, options = {})
        post = initial_post(options)
        add_creditcard(post, creditcard)

        commit('credit_return', money, post)
      end
      alias refund return

      def reverse(money, authorization, options = {})
        post = initial_post(options)
        add_authorization(post, authorization)

        commit('credit_reversal', money, post)
      end

      def void(authorization, options = {})
        post = initial_post(options)
        add_authorization(post, authorization)

        commit('credit_void', nil, post)
      end

      private

      def add_customer_data(post, options)
        # TODO?
      end

      def add_address(post, creditcard, options)
        address = options[:billing_address] || {}
        post[:card_holder_data] = {
          :card_holder_first_name => creditcard.first_name,
          :card_holder_last_name  => creditcard.last_name,
          :card_holder_addr       => [address[:address1], address[:address2]].compact.join(" "),
          :card_holder_city       => address[:city],
          :card_holder_state      => address[:state],
          :card_holder_zip        => address[:zip].to_s.gsub(/[^0-9A-Za-z]/, ''),
          :card_holder_phone      => address[:phone].to_s.gsub(/[^0-9]/, ''),
          #:card_holder_email     => "",
        }
      end

      def add_authorization(post, authorization)
        post[:gateway_txn_id] = authorization
      end

      def add_creditcard(post, creditcard)
        post[:card_data] = {
          :manual_entry => {
            :card_nbr  => creditcard.number,
            :exp_month => creditcard.month,
            :exp_year  => creditcard.year,
            :c_v_v_2   => creditcard.verification_value,
            :card_present   => 'N',
            :reader_present => 'N',
          }
        }
      end

      def initial_post(options)
        # Not usually expected to be true, mostly just here for
        # testing purposes
        { :allow_dup => options[:allow_dup] }
      end

      def add_invoice(post, options)
        # avoid empty hash
        return unless options[:invoice] || options[:invoice_ship_month] || options[:invoice_ship_day]
        post[:direct_mkt_data] = {
          :direct_mkt_invoice_nbr => options[:invoice],
          :direct_mkt_ship_month  => options[:invoice_ship_month],
          :direct_mkt_ship_day    => options[:invoice_ship_day],
        }
      end

      def client
        @client ||= HeartlandPortico.new(@options, test?)
      end

      def commit(action, money, parameters)
        parameters[:amt] = amount(money)
        translate_response(client.send(action, parameters))
      end

      def translate_response(heartland_response)
        Response.new(
          heartland_response.success?,
          heartland_response.message,
          heartland_response.body,
          { :fraud_review => false,
            :authorization => heartland_response.authorization,
            :test => test?,
            :avs_result => { :code => heartland_response.avs_code },
            :cvv_result => heartland_response.cvv_code
          }
        )
      end
    end
  end
end
